@file:JvmName("TransactionMethodsKt")
@file:OptIn(ExperimentalContracts::class, LowLevelApi::class)
@file:Suppress("FunctionName")

package ru.landgrafhomyak.bgtu.db0.storage.serdha.schema._temp

import kotlin.contracts.ExperimentalContracts
import kotlin.contracts.InvocationKind
import kotlin.contracts.contract
import kotlin.jvm.JvmName
import ru.landgrafhomyak.db.serdha0.api.LowLevelApi
import ru.landgrafhomyak.db.serdha0.api.queries._Query
import ru.landgrafhomyak.db.serdha0.api.runtime.InputRow
import ru.landgrafhomyak.db.serdha0.api.runtime.OutputRow
import ru.landgrafhomyak.db.serdha0.api.runtime.Transaction


@PublishedApi
internal inline fun <R> _safeAutoClose(onAbort: () -> Unit = {}, onSuccess: () -> Unit = {}, action: () -> R): R {
	contract {
		callsInPlace(action, InvocationKind.EXACTLY_ONCE)
		callsInPlace(onAbort, InvocationKind.AT_MOST_ONCE)
		callsInPlace(onSuccess, InvocationKind.AT_MOST_ONCE)
	}
	val ret: R
	try {
		ret = action()
	} catch (e1: Throwable) {
		try {
			onAbort()
		} catch (e2: Throwable) {
			e1.addSuppressed(e2)
		}
		throw e1
	}
	onSuccess()
	return ret
}

class ExpectedOneRowException() : Exception()
class TooManyRowsException() : Exception()

suspend inline fun <_qUE : Any, R> Transaction.selectOneOrError(
	compiledQuery: _Query.Params2Table<_qUE>,
	params: (qUE: _qUE, i: InputRow<_qUE>) -> Unit,
	result: (qUE: _qUE, o: OutputRow<_qUE>) -> R
): R {
	val i = this._executeQuery(compiledQuery)
	_safeAutoClose(onAbort = { i._abort() }) {
		params(compiledQuery.uExt, i)
	}
	val o = i._finish()
	val ret: R
	_safeAutoClose(onAbort = { o._abort() }, onSuccess = { o._finish() }) {
		if (!o._next())
			throw ExpectedOneRowException()
		ret = result(compiledQuery.uExt, o)
		if (o._next())
			throw TooManyRowsException()
	}
	return ret

}
